# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from sage.rings.rational_field import QQ
from sage.rings.complex_mpfr import ComplexField
from sage.graphs.graph import Graph
from sage.rings.imaginary_unit import I
from sage.functions.other import arg

from sage.misc.flatten import flatten

from sage.graphs.spanning_tree import boruvka
from delaunay_triangulation import triangulate

import os

from .util import Util

class FundamentalGroupDelaunayDual(object):
    def __init__(self, points, basepoint, border=5):
        assert basepoint not in points

        self._points = [basepoint] + points
        self._border = border

        self.CC = ComplexField(50) # ultimately this should be dropped for certified precision


    def rationalize(self, z):
        zcc = self.CC(z)
        zr, zi = zcc.real(), zcc.imag()
        zq = Util.simple_rational(zr, self.prec) + I*Util.simple_rational(zi, self.prec)
        return zq

    def point_to_complex_number(self, pt):
        return pt[0] + I*pt[1]
    def complex_number_to_point(self, z):
        return (QQ(z.real()), QQ(z.imag()))

    @property
    def prec(self):
        if not hasattr(self, "_prec"):
            points_differences = flatten([[self.CC(self._points[i]) - self.CC(self._points[j]) for j in range(i)] for i in range(len(self._points))])
            self._prec = Util.simple_rational(min([abs(p) for p in points_differences]), min([abs(p) for p in points_differences])/100)/100
        return self._prec
    
    @property
    def points(self):
        return self._points
    

    @property
    def vertices(self):
        if not hasattr(self, "_vertices"):
            polygons = self.polygons 
            # calling self.polygons defines self._vertices -- this prevents unecessary 
            # loops, but there is likely a better way to achieve the same goal
        return self._vertices
    
    @property
    def duality(self):
        if not hasattr(self, "_duality"):
            polygons = self.polygons 
            # calling self.polygons defines self._duality -- this prevents unecessary 
            # loops, but there is likely a better way to achieve the same goal
        return self._duality

    @property
    def edges(self):
        """ returns the edges of the Voronoi graph given as pairs of indices of self.vertices, in decreasing length
        """
        if not hasattr(self, "_edges"):
            edges = []
            for center, polygon in self.polygons:
                for e in polygon:
                    if e not in edges and [e[1], e[0]] not in edges:
                        edges += [e]
            connection_to_basepoint = min([i for i in range(1, len(self.vertices))], key=lambda i: abs(self.vertices[0] - self.vertices[i]))
            edges += [[0, connection_to_basepoint]]
            edges.sort(reverse=True, key=lambda e:(self.vertices[e[0]].real()-self.vertices[e[1]].real())**2 + (self.vertices[e[0]].imag()-self.vertices[e[1]].imag())**2)
            self._edges = edges
        return self._edges

    @property
    def border(self):
        return self._border

    @property
    def qpoints(self):
        if not hasattr(self, "_qpoints"):
            self._qpoints = [self.rationalize(p) for p in self.points]
        return self._qpoints


    @property
    def graph(self):
        if not hasattr(self, "_graph"):
            self._graph = Graph([(e[0], e[1], self.rationalize(abs(self.vertices[e[0]] - self.vertices[e[1]]))) for e in self.edges])
        return self._graph

    @property
    def tree(self):
        if not hasattr(self, "_tree"):
            tree_edges = boruvka(self.graph)
            self._tree = Graph(tree_edges)
        return self._tree

    @property
    def loop_points(self):
        if not hasattr(self, "_loop_points"):
            loop_points = []
            for i, loop in enumerate(self.loops):
                loop_point = min(loop[1:], key=lambda v:self.tree.distance(v,0))
                index = loop.index(loop_point)
                if index!=0:
                    loop = loop[index:-1] + loop[:index] + [loop[index]]
                self.loops[i] = loop
                loop_points+=[loop_point]
            self._loop_points = loop_points
        return self._loop_points

    @property
    def paths(self):
        if not hasattr(self, "_paths"):
            self._paths = [self.tree.all_paths(0, v)[0] for v in self.loop_points]
        return self._paths
    
    @property
    def pointed_loops(self):
        if not hasattr(self, "_pointed_loops"):
            pointed_loops = []
            for i in range(len(self.points)-1):
                pointed_loops += [self.paths[i][:-1] + self.loops[i] + list(reversed(self.paths[i][:-1]))]
            self._pointed_loops = pointed_loops
        return self._pointed_loops
    
    

    @property
    def loops(self):
        if not hasattr(self, "_loops"):
            loops = []
            for center, polygon in self.polygons:
                polygon = [[v for v in e] for e in polygon]
                loop = polygon.pop()
                while len(polygon) > 0:
                    for i in range(len(polygon)):
                        e = polygon[i]
                        if e[0] == loop[-1]:
                            loop += [e[1]]
                        elif e[1] == loop[-1]:
                            loop += [e[0]]
                        else:
                            assert i!= len(polygon), "polygon is not a loop, possibly because the voronoi cell is not bounded"
                            continue
                        polygon.pop(i)
                        break

                loops += [list(reversed(loop))] if Util.is_clockwise([self.vertices[i] for i in loop[:-1]]) else [loop]
            self._loops = loops
        return self._loops

    @property
    def minimal_tree(self):
        if not hasattr(self, "_minimal_tree"):
            edges = []
            for path in self.paths:
                for i in range(len(path)-1):
                    e0 = path[i]
                    e1 = path[i+1]
                    if [e0,e1] not in edges and [e1,e0] not in edges:
                        edges+=[[e0, e1]]
            self._minimal_tree = Graph(edges)
        return self._minimal_tree

    def neighbours(self, v): # maybe this shoudl be done only once?
        neighbours = self.graph.neighbors(v)
        neighbours.sort(key=lambda v2:arg(self.vertices[v2] - self.vertices[v]))
        return neighbours

    def sort_loops(self):
        order = self._sort_loops_rec(0)
        self._points = [self.points[0]] + [self.points[i+1] for i in order]
        self._qpoints = [self.qpoints[0]] + [self.qpoints[i+1] for i in order]
        self._rootapprox = self._qpoints + self._rootapprox[len(self._qpoints):]
        self._loops = [self.loops[i] for i in order]
        self._paths = [self.paths[i] for i in order]
        self._duality = [[edge, [order.index(i-1)+1 if i!=0 and i<len(self.points) else i for i in dual]] for [edge, dual] in self._duality]
        if hasattr(self, "_pointed_loops"):
            self._pointed_loops = [self.pointed_loops[i] for i in order]
        return order

    def adapted_loops(self, subvoronoi): # This is broken.
        for v in subvoronoi.points:
            assert v in self.points
        correspondance = []
        for v in subvoronoi.vertices:
            correspondance += [Util.select_closest_index(self.vertices,v)]
        adapted_loops = []
        for loop in subvoronoi.pointed_loops:
            adapted_loop = [correspondance[loop[0]]]
            for v in loop[1:]:
                adapted_loop += self.graph.shortest_path(adapted_loop[-1], correspondance[v])[1:]
            adapted_loops += [adapted_loop]
        return adapted_loops


    def _sort_loops_rec(self, v, parent=None, depth=0):
        neighbours = self.neighbours(v)
        tree_neighbours = self.minimal_tree.neighbors(v)
        loops = [(i, loop[1]) for i, loop in enumerate(self.loops) if loop[0]==v]
        if parent!=None:
            index = neighbours.index(parent)
            neighbours = neighbours[index:] + neighbours[:index]

        order = []
        for child in neighbours:
            if child!= parent and child in tree_neighbours:
                order+=self._sort_loops_rec(child, v, depth+1)
            for loop in loops:
                if loop[1] == child:
                    order+=[loop[0]]
        return order


    @property
    def polygons(self): # this interfacing with VoronoiDiagram is so ugly
        if not hasattr(self, "_polygons"):
            vertices = [self.points[0]]
            rootapprox = [p for p in self.qpoints] # there should be a `copy` function

            qpoints = [self.complex_number_to_point(z) for z in self.qpoints]
            reals = [s[0] for s in qpoints]
            imags = [s[1] for s in qpoints]
            xmin, xmax, ymin, ymax = min(reals), max(reals), min(imags), max(imags)
            shift = max(ymax-ymin, xmax-xmin)/2 # there is likely something more clever to do here
            xmax, ymax = max(xmax, ymax),max(xmax, ymax)
            xmin, ymin = min(xmin, ymin),min(xmin, ymin)
            xmin, xmax, ymin, ymax = xmin - shift, xmax + shift, ymin - shift, ymax + shift
            
            for i in range(self.border):
                step = QQ(i)/QQ(self.border)
                rootapprox += [xmin + step*(xmax-xmin) + I*ymax]
                rootapprox += [xmax + step*(xmin-xmax) + I*ymin]
                rootapprox += [xmin + I*(ymin + step*(ymax-ymin))]
                rootapprox += [xmax + I*(ymax + step*(ymin-ymax))]
            
            list_of_rational_points = [] # why is this necessary? I do not know, but otherwise VoronoiDiagram throws a fit
            for z in rootapprox:
                p = self.complex_number_to_point(z)
                list_of_rational_points+= [[QQ(p[0]), QQ([p[1]])]]

            delvertices = [triangulate.Vertex(p.real(), p.imag()) for p in rootapprox]
            dt = triangulate.delaunay(delvertices)

            self._delaunay_triangulation = dt
            self._rootapprox = rootapprox
            
            polygons_temp = [[p, []] for p in rootapprox]
            
            for triangle in dt:
                ts = [t.x+I*t.y for t in triangle]
                center = sum(ts)/3
                for i1, i2 in [[0,1],[1,2], [2,0]]:
                    p1 = ts[i1]
                    p2 = ts[i2]
                    middle = (p1+p2)/2
                    edge = [center, middle]

                    v1, v2 = Util.select_closest_index(rootapprox, p1), Util.select_closest_index(rootapprox, p2)
                    

                    for v in [v1, v2]:
                        polygons_temp[v][1]+=[[edge, [v1,v2]]]

            # we are only interested in cells around elements of self.points
            indices = [Util.select_closest_index([center for center, polygon in polygons_temp], c) for c in self.qpoints[1:]]
            polygons_temp = [polygons_temp[i] for i in indices]

            # then we translate the edges in rational coordinate as well
            polygons = []
            duality = []
            all_edges = []
            for center, polygon in polygons_temp:
                edges = []
                for edge, dual in polygon:
                    e0 = self.rationalize(edge[0])
                    if e0 not in vertices:
                        vertices += [e0]
                    e1 = self.rationalize(edge[1])
                    if e1 not in vertices:
                        vertices += [e1]
                    if e0 != e1:
                        e = [vertices.index(e0),vertices.index(e1)]
                        edges += [e]
                        if e not in all_edges and (dual[0] < len(self.points) and dual[1] < len(self.points)):
                            duality += [[e, dual]]
                            all_edges += [e]
                polygons += [[center, edges]]

            self._vertices = vertices
            self._polygons = polygons
            self._duality = duality

        return self._polygons








# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from .numperiods.family import Family
from .numperiods.integerRelations import IntegerRelations
from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.qqbar import QQbar
from sage.rings.integer_ring import ZZ
from sage.rings.number_field.number_field import NumberField
from sage.rings.imaginary_unit import I
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.functions.other import factorial
from sage.functions.other import floor
from sage.matrix.constructor import matrix
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix
from sage.arith.functions import lcm
from sage.arith.misc import gcd
from sage.arith.misc import factor
from sage.symbolic.constants import pi
from sage.functions.transcendental import zeta
from sage.misc.misc_c import prod
from sage.groups.matrix_gps.symplectic import Sp

from sage.misc.flatten import flatten
from sage.misc.functional import log

from .voronoi import FundamentalGroupVoronoi
from .integrator import Integrator
from .util import Util
from .context import Context

import logging
import time

logger = logging.getLogger(__name__)


class CalabiYauOperator(object):
    def __init__(self, L, basepoint=None, **kwds):
        """L, a differential equation 
        """
        
        self.ctx = Context(**kwds)

        self._L = L
        
        if basepoint!= None: # it is useful to be able to specify the basepoint to avoid being stuck in arithmetic computations if critical values have very large modulus
            assert basepoint not in self.singular_values, "basepoint is not regular"
            self._basepoint = basepoint

        if not self.ctx.debug:
            fg = self.fundamental_group # this allows reordering the critical points straight away and prevents shenanigans. There should be a better way to do this
    
    @property
    def L(self):
        """returns the Calabi-Yau operator"""
        return self._L
    
    @property
    def order(self):
        return self.L.order()
    
    @property
    def basepoint(self):
        if not hasattr(self, "_basepoint"):
            self._basepoint = QQ(10)**(min([floor(log(abs(r),10)) for r in self.singular_values if r!=0])-1)
        return self._basepoint
   
    @property
    def singular_values(self):
        if not hasattr(self, "_singular_values"):
            self._singular_values = self.L.leading_coefficient().roots(QQbar, multiplicities=False)
        return self._singular_values
 
    @property
    def fundamental_group(self):
        if not hasattr(self, "_fundamental_group"):
            fundamental_group = FundamentalGroupVoronoi(self.singular_values, self.basepoint)
            self._singular_values = [self._singular_values[i] for i in fundamental_group.sort_loops()]
            self._fundamental_group = fundamental_group
        return self._fundamental_group
    

    @property
    def maximal_unipotent_monodromy_index(self):
        """returns the Calabi-Yau operator"""
        return self._fundamental_group.points.index(0)-1 # TODO : what if we want a different MUM_point ?



    @property
    def transition_matrices(self):
        if not hasattr(self, "_transition_matrices"):
            integrator = Integrator(self.fundamental_group, self.L, self.ctx.nbits)
            transition_matrices = integrator.transition_matrices
            if self.L.annihilator_of_composition(1/self.L.base_ring().gen()).leading_coefficient()(0)==0:
                transition_matrices += [prod(list(reversed(transition_matrices))).inverse()]
                self._singular_values = self.singular_values + ["infinity"]
            self._transition_matrices = transition_matrices
        return self._transition_matrices
    
    @property
    def basis_change_to_mum_point(self):
        if not hasattr(self, "_basis_change_to_mum_point"): # performance can be gained here (redundant computation with self.transition_matrices)
            path = [self.fundamental_group.vertices[i] for i in self.fundamental_group.paths[self.maximal_unipotent_monodromy_index]] + [0]
            self._basis_change_to_mum_point = self.L.numerical_transition_matrix(path, ZZ(2)**-self.ctx.nbits, assume_analytic=True)
        return self._basis_change_to_mum_point
    

    def _discover_rational_basis(self, starting_vector):
        vs = [starting_vector]
        while len(vs)!=4:
            found = False
            for M in self.transition_matrices:
                for v in vs:
                    Pi = matrix(vs + [M*v]).transpose()
                    if any([m!=0 for m in Pi.minors(Pi.ncols())]):
                        vs += [M*v]
                        found = True
            if not found:
                self._is_semi_simple = False
                break
        if found:
            self._is_semi_simple = True
        return matrix(vs).transpose()

    @property
    def period_matrix(self):
        if not hasattr(self, "_period_matrix"):
            Mmum = self.transition_matrices[self.maximal_unipotent_monodromy_index]
            for starting_vector in ((Mmum-1)**3).columns():
                if starting_vector!=0:
                    break
            Pi = self._discover_rational_basis(starting_vector)
            for generator_MUM in Pi.columns():
                if (self.basis_change_to_mum_point * generator_MUM)[0] != 0:
                    break
            Pi = matrix([(Mmum-1)**k*generator_MUM for k in [0,1,3,2]]).transpose()

            Pi = Pi * Util.saturate([(Pi.inverse() * M.inverse() * Mmum * M * Pi).apply_map(Util.rationalize) for M in self.transition_matrices])

            self._period_matrix =  Pi 
        return self._period_matrix

    def rationalise_periods(self, periods):
        twopiI = self.ctx.CBF(ZZ(2)*pi*I)
        finalCB = (diagonal_matrix([1, twopiI**-1, twopiI**-2, twopiI**-3]) * matrix(ZZ, 4, 4, {(0,3):1, (1,2):1, (2,1):1, (3,0):1}) * diagonal_matrix([6,2,1,1])).transpose()
        period_matrix = self.basis_change_to_mum_point * periods

        
        M = (finalCB.inverse()*period_matrix).transpose()

        SF, r = self.number_field
        nu = SF.gen(0)
        R = PolynomialRing(SF, 'lambda3')
        constants = [1, self.ctx.CBF(zeta(3))/twopiI**3]
        variables = [1, R.gen(0)]

        constantsall, variablesall = [], []
        for i in range(SF.degree()):
            constantsall += [self.ctx.CBF(r)**i * v for v in constants]
            variablesall += [nu**i * v for v in variables]

        formal_matrix = M.apply_map(lambda c: Util.get_coefficient(c, constantsall, variablesall))

        return formal_matrix





    @property
    def number_field(self):
        if not hasattr(self, "_number_field"):
            for starting_vector in ((self.transition_matrices[self.maximal_unipotent_monodromy_index]-1)**3).columns():
                if starting_vector!=0:
                    break
            Pi = self._discover_rational_basis(starting_vector)
            t = self.L.base_ring().gen(0)
            coefs = flatten([(Pi.inverse() * M * Pi).coefficients() for M in self.transition_matrices])
            unknowns = [1]
            for c in coefs:
                if IntegerRelations(matrix(unknowns+[c]).transpose()).basis.nrows()==0:
                    unknowns += [c]

            self._has_integral_monodromy = len(unknowns) == 1

            if not self.has_integral_monodromy:
                P = lcm([sum([c*t**i for i, c in enumerate(Util.check_if_algebraic(u))]) for u in unknowns[1:]])
                SF = P.splitting_field("nu").optimized_representation(name="nu")[0]
                P = SF.defining_polynomial()

                r = P.roots(QQbar, multiplicities=False)[1]
                SF = NumberField(P,"nu", embedding=r)
                self._number_field = (SF, r)
            else: 
                self._number_field = (QQ, 1)
        return self._number_field
    

    @property # todo: get intersection product from top invariants and check it is correct (i.e. invariant under monodromy).
    def intersection_product(self):
        Mmum = self.transition_matrices[self.maximal_unipotent_monodromy_index]
        Pi = self.period_matrix
        Ms = [(Pi.inverse() * M.inverse() * Mmum * M * Pi).apply_map(Util.rationalize) for M in self.transition_matrices]
        return self.compute_intersection_product(Ms)

    def compute_intersection_product(self, Ms):
        R = PolynomialRing(QQ, self.order**2, 'a')
        IP = matrix(self.order,self.order,R.gens())
        I = R.ideal(flatten((IP+IP.transpose()).coefficients()+[(M.transpose()*IP*M-IP).coefficients() for M in Ms]))
        gens = I.groebner_basis()
        for a in R.gens():
            if a not in gens:
                break
        I2 = R.ideal(flatten([a-1] + (IP+IP.transpose()).coefficients()+[(M.transpose()*IP*M-IP).coefficients() for M in Ms]))
        IP = IP.apply_map(lambda c: I2.reduce(c)).change_ring(QQ)
        IP = IP * lcm([c.denominator() for c in IP.coefficients()])
        IP = IP.change_ring(ZZ)
        return IP

    @property
    def types(self):
        if not hasattr(self, "_types"):
            types = [self._monodromy_type(M) for M in self.transition_matrices]
            self._types = types
        return self._types
    
    @classmethod
    def _monodromy_type(cls, M):
        for i in range(1,13):
            for j in range(1,5):
                try:
                    if ((M**i-1)**j).change_ring(ZZ) == 0:
                        return (i,j)
                except KeyboardInterrupt:
                    raise KeyboardInterrupt
                except:
                    continue
        raise Exception("Monodromy type not identified")

    @property
    def maximal_unipotent_monodromy_fibres(self):
        if not hasattr(self, "_maximal_unipotent_monodromy_fibres"):
            self._maximal_unipotent_monodromy_fibres = [i for i, t in enumerate(self.types) if t == (1,4)]
        return self._maximal_unipotent_monodromy_fibres
    
    @property
    def conifold_fibres(self):
        if not hasattr(self, "_conifold_fibres"):
            conifold_fibres = [i for i, t in enumerate(self.types) if t == (1,2)]
            conifold_fibres.sort(key = lambda i: abs(self.ctx.CBF(self.singular_values[i]))if self.singular_values[i]!="infinity" else max([abs(self.ctx.CBF(c)) for c in self.singular_values if c!="infinity"])+1)
            self._conifold_fibres = conifold_fibres
        return self._conifold_fibres


    def _extend(self, vs, Ms):
        CB = matrix(vs).transpose()
        newMs = [(CB.inverse()*M*CB) for M in Ms]
        if self.has_integral_monodromy:
            newCB = CB*Util.saturate(newMs)
            newMs = [(newCB.inverse()*M*newCB).change_ring(ZZ) for M in Ms]
        IP = self.compute_intersection_product(newMs).change_ring(ZZ)
        return IP.det()

    @property
    def gamma_class(self):
        if not hasattr(self, "_gamma_class"):
            gamma_class = self.compute_decomposition_in_mum_frobenius(self.period_matrix * self.standard_basis)
            self._gamma_class = gamma_class
        return self._gamma_class
    
    @property
    def monodromy_matrices(self):
        if not hasattr(self, "_monodromy_matrices"):
            standard_periods = self.period_matrix * self.standard_basis
            self._monodromy_matrices = self.monodromy_from_periods(standard_periods)
        return self._monodromy_matrices

    @property
    def laddered_basis(self):
        if not hasattr(self, "_laddered_basis"):
            monodromy_matrices = self.monodromy_from_periods(self.period_matrix)
            Mmum = monodromy_matrices[self.maximal_unipotent_monodromy_index]
            holo = ((Mmum-1)).right_kernel().gen(0)

            found = False
            for i in self.conifold_fibres:
                M = monodromy_matrices[i]
                if M*holo not in ((Mmum-1)**3).right_kernel():
                    log4 = M*holo
                    found = True
                    break

            if not found:
                for M in monodromy_matrices:
                    if M*holo not in ((Mmum-1)**3).right_kernel():
                        log4 = M*holo
                        found = True
                        break
            if found:
                d = (2*3*5*7*11*13*17*19*21*23*29)**10
                d = 1
                holo, log2, log3 = (Mmum-1)**3*log4/d, (Mmum-1)**2*log4/d, (Mmum-1)*log4

            if not found:
                vs = [holo]
                log2ker =((Mmum-1)**2).right_kernel_matrix()
                log2 = vector(Util.find_complement(log2ker.solve_left(matrix(vs)), primitive=False) * log2ker)
                vs += [log2]
                
                log3ker =((Mmum-1)**3).right_kernel_matrix()
                log3 = vector(Util.find_complement(log3ker.solve_left(matrix(vs)), primitive=False) * log3ker)
                vs += [log3]
                
                log4 = vector(Util.find_complement(matrix(vs), primitive=False))
                vs += [log4]

            vs = [log4, log3, log2, holo]
            vs = matrix(vs).transpose()
            
            CB = Util.saturate([vs.inverse() * M * vs for M in monodromy_matrices]) if self.has_integral_monodromy else identity_matrix(4)
            perm = matrix(ZZ, 4, 4, {(0,0):1, (1,1):1, (2,3):1, (3,2):1})
            self._laddered_basis = (vs * CB * perm)
        return self._laddered_basis
    

    def normalize_basis(self, basis):
        monodromy_matrices = self.monodromy_from_periods(self.period_matrix*basis)
        if self.has_integral_monodromy:
            IP = self.compute_intersection_product(monodromy_matrices)
            vs = identity_matrix(4).rows()
            vs[3] -= vs[2] * ((vs[0] * IP * vs[3])//(vs[0] * IP * vs[2]))
            vs[1] -= vs[2] * ((vs[0] * IP * vs[1])//(vs[0] * IP * vs[2]))
            vs = matrix(vs).transpose()

            res = self.compute_decomposition_in_mum_frobenius(self.period_matrix * basis * vs)

            vs = vs.columns()
            vs[3] -= ((QQ(res[3,0]).numerator() * QQ(res[2,0]).denominator()) // (QQ(res[2,0]).numerator() * QQ(res[3,0]).denominator()))*vs[2]
            vs[0] -= vs[1] * ((vs[0] * IP * vs[3])//(vs[1] * IP * vs[3]))
            vs[3] -= vs[2] * ((vs[0] * IP * vs[3])//(vs[0] * IP * vs[2]))

            if res[3,1]<0:
                vs[3] = -vs[3]
            if vs[3] * IP * vs[1] >0:
                vs[1] = -vs[1]
            if res[2,0]<0:
                vs[2] = -vs[2]
            if vs[2] * IP * vs[0] >0:
                vs[0] = -vs[0]
            
            vs = matrix(vs).transpose()

            res = self.compute_decomposition_in_mum_frobenius(self.period_matrix * basis * vs)

            vs = vs.columns()
            vs[0] -= floor(res[0,0].monomial_coefficient(res[0,0].parent()(1))) * vs[2]
            vs[1] -= (QQ(res[1,1]).numerator() * QQ(res[3,1]).denominator()) // (QQ(res[3,1]).numerator() * QQ(res[1,1]).denominator()) * vs[3]
            vs = matrix(vs).transpose()

            return basis * vs
        return basis
    
    def monodromy_from_periods(self, periods):
        SF, r = self.number_field
        nu, d = SF.gen(0), SF.degree()
        variables = [r**i for i in range(d)]
        constants = [nu**i for i in range(d)]
        return [(periods.inverse() * M * periods).apply_map(lambda c: Util.get_coefficient(c, variables, constants)) for M in self.transition_matrices]

    @property
    def standard_basis(self):
        perm = matrix(ZZ, 4, 4, {(0,0):1, (1,1):1, (2,3):1, (3,2):1})
        basis = self.normalize_basis(self.laddered_basis) * perm

        monodromy_matrices = self.monodromy_from_periods(self.period_matrix * basis)
        CB = identity_matrix(4)
        if self.has_integral_monodromy: #the following is outdated 
            # vs = self._reduce_discriminant(identity_matrix(4).rows(), monodromy_matrices)
            vs = identity_matrix(4)
            CB = matrix(vs).transpose()
            CB = CB * Util.saturate([CB.inverse() * M * CB for M in monodromy_matrices])

        basis = self.normalize_basis(basis * CB * perm)

        basis = basis * lcm([c.denominator() for c in basis.coefficients()])

        return basis


    def compute_decomposition_in_mum_frobenius(self, periods):
        twopiI = self.ctx.CBF(ZZ(2)*pi*I)
        finalCB = (diagonal_matrix([1, twopiI**-1, twopiI**-2, twopiI**-3]) * matrix(ZZ, 4, 4, {(0,3):1, (1,2):1, (2,1):1, (3,0):1}) * diagonal_matrix([6,2,1,1])).transpose()
        period_matrix = self.basis_change_to_mum_point * periods

        Mmum = (self.monodromy_from_periods(periods)[self.maximal_unipotent_monodromy_index]-1)
        if self.has_integral_monodromy:
            Mmum = Mmum.change_ring(ZZ)
        c = period_matrix * Mmum.right_kernel().gen(0)
        period_matrix = period_matrix / c[-1]
        M = (finalCB.inverse()*period_matrix).transpose()

        
        R = PolynomialRing(QQ, 'lambda3')
        constants = [1, self.ctx.CBF(zeta(3))/twopiI**3]
        variables = [1, R.gen(0)]

        try:
            formal_matrix = M.apply_map(lambda c: Util.get_coefficient(c, constants, variables))
            return formal_matrix
        except:
            pass
        
        SF, r = self.number_field
        nu = SF.gen(0)
        R = PolynomialRing(SF, 'lambda3')
        constants = [1, self.ctx.CBF(zeta(3))/twopiI**3]
        variables = [1, R.gen(0)]

        constantsall, variablesall = [], []
        for i in range(SF.degree()):
            constantsall += [self.ctx.CBF(r)**i * v for v in constants]
            variablesall += [nu**i * v for v in variables]

        formal_matrix = M.apply_map(lambda c: Util.get_coefficient(c, constantsall, variablesall))

        return formal_matrix
    

    @property
    def monodromy_in_scaled_frobenius_basis(self):
        twopiI = self.ctx.CBF(ZZ(2)*pi*I)
        scaled_frobenius_basis = (diagonal_matrix([1, twopiI**-1, twopiI**-2, twopiI**-3]) * matrix(ZZ, 4, 4, {(0,3):1, (1,2):1, (2,1):1, (3,0):1}) * diagonal_matrix([6,2,1,1])).transpose()
        period_matrix = self.basis_change_to_mum_point.inverse() * scaled_frobenius_basis

        monodromy_matrices = [period_matrix.inverse() * M * period_matrix for M in self.transition_matrices]
        
        SF, r = self.number_field
        nu = SF.gen(0)
        R = PolynomialRing(SF, 'lambda')
        l = R.gen(0)
        variables = [1, l, l**2]
        hom = R.hom([self.ctx.CBF(zeta(3))/twopiI**3])
        constants = [hom(v) for v in variables]

        constantsall, variablesall = [], []
        for i in range(SF.degree()):
            constantsall += [self.ctx.CBF(r)**i * v for v in constants]
            variablesall += [nu**i * v for v in variables]

        monodromy_matrices = [M.apply_map(lambda c: Util.get_coefficient(c, constantsall, variablesall)) for M in monodromy_matrices]

        return monodromy_matrices
    

    @property
    def is_semi_simple(self):
        """returns whether the monodromy group is semi-simple"""
        if not hasattr(self, "_is_semi_simple"):
            for starting_vector in ((self.transition_matrices[self.maximal_unipotent_monodromy_index]-1)**3).columns():
                if starting_vector!=0:
                    break
            self._discover_rational_basis(starting_vector)
        return self._is_semi_simple
    
    @property
    def has_conifold(self):
        if not hasattr(self, "_has_conifold"):
            self._has_conifold = len(self.conifold_fibres)>0
        return self._has_conifold
    
    @property
    def is_symplectic(self):
        if not hasattr(self, "_is_symplectic"):
            self._is_symplectic = self.intersection_product == matrix(ZZ, 4, 4, {(0,2):1, (1,3):1, (2,0):-1, (3,1):-1})
        return self._is_symplectic
    
    @property
    def has_two_maximal_unipotent_monodromy_points(self):
        if not hasattr(self, "_has_two_maximal_unipotent_monodromy_points"):
            self._has_two_maximal_unipotent_monodromy_points = len(self.maximal_unipotent_monodromy_fibres)>1
        return self._has_two_maximal_unipotent_monodromy_points
    
    @property
    def has_integral_monodromy(self):
        if not hasattr(self, "_has_integral_monodromy"):
            self.number_field
        return self._has_integral_monodromy

    @property
    def canonical_topological_invariants(self):
        """ returns (H3, c2H, chi, p) where H3 is the triple intersection number, c2H is the second Chern class, chi is the top Chern class (the Euler characteristic), """
        raise NotImplemented

    @property
    def c2H(self):
        """returns the second Chern class of the operator"""
        raise NotImplemented

    @property
    def triple_intersection(self):
        """returns the triple intersection number of the operator"""
        raise NotImplemented

    @property
    def paths(self):
        if not hasattr(self, "_paths"):
            paths = self.fundamental_group.pointed_loops_vertices
            self.transition_matrices
            if "infinity" in self.singular_values:
                paths += [-sum(paths)]
            self._paths = paths
        return self._paths
    
    def monodromy_index_mod_n(self, n):
        if not hasattr(self, "_reduction_indices"):
            self._reduction_indices = {}
        res = 1
        for p, power in factor(n):
            q = p**power
            if q not in self._reduction_indices.keys():
                self._reduction_indices[q] = self._monodromy_index_mod_q(q)
            res *= self._reduction_indices[q]
        return res

    def _monodromy_index_mod_q(self, q):
        R = IntegerModRing(q)
        monodromy_matrices_mod_q = [M.change_ring(R) for M in self.monodromy_matrices]
        Sp4 = Sp(4, R)
        if Sp4.invariant_form() == matrix(R, 4, 4, {(0,3):1, (1,2):1, (2,1):-1, (3,0):-1}):
            CB = matrix(R, 4, 4, {(0,0):1, (1,1):1, (2,3):1, (3,2):1})
            monodromy_matrices_mod_q = [CB.inverse()*M*CB for M in monodromy_matrices_mod_q]
        monodromy_group = Sp4.subgroup(monodromy_matrices_mod_q)
        return Sp4.cardinality()/monodromy_group.cardinality()


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all


from .numperiods.family import Family
from .numperiods.integerRelations import IntegerRelations
from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.qqbar import QQbar
from sage.functions.other import factorial
from sage.functions.other import floor
from sage.matrix.constructor import matrix
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix
from sage.arith.functions import lcm


from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.schemes.toric.weierstrass import WeierstrassForm
from sage.misc.flatten import flatten

from sage.modules.free_quadratic_module_integer_symmetric import IntegralLattice

from sage.functions.other import ceil

from .voronoi import FundamentalGroupVoronoi
from .integrator import Integrator
from .util import Util
from .context import Context
from .monodromyRepresentation import MonodromyRepresentation
from .monodromyRepresentationFiberedProduct import MonodromyRepresentationFibreProduct
from sage.functions.other import binomial


import logging
import time

logger = logging.getLogger(__name__)

def tens1(M):
    return block_diagonal_matrix([M, M], subdivide=False)
def tens2(M):
    return block_matrix(list(map(lambda r:list(map(lambda c: c*identity_matrix(M.nrows()), r)), M.rows())), subdivide=False)
def tensor_prod(a,b):
    return vector([a[0]*b[0], a[1]*b[0], a[0]*b[1], a[1]*b[1]])

def derivatives_coords(S, w, k):
    derivatives = [S.P.parent()(0), w]
    for k in range(k):
        derivatives += [S._derivative(derivatives[-1], S.P)] 
    return S.family._coordinates(derivatives, S.basepoint)

def derivative_of_product(a, b, n):
    return sum([binomial(n, i) * tensor_prod(a.row(i+1), b.row(n-i+1)) for i in range(n+1)])

class FibreProduct(object):
    def __init__(self, S1, S2, correction=None, **kwds) -> None:
        """S1 and S2 are two elliptic surfaces with the same basepoint
        """
        
        self.ctx = Context(**kwds)
        
        self._S1 = S1
        self._S2 = S2

        if correction!=None:
            self._correction = correction

        assert S1.basepoint == S2.basepoint, "the basepoint of the input elliptic surfaces should be the same"
        self._basepoint=S1.basepoint
    
    @property
    def correction(self):
        if not hasattr(self, "_correction"):
            pmin=None
            L1 = self.S1.family.picard_fuchs_equation(vector([1,0]))
            L2 = self.S2.family.picard_fuchs_equation(vector([1,0]))
            L = L1.symmetric_product(L2)
            for v in L.local_basis_monomials(0):
                if "log" in str(v):
                    continue
                if str(v)=="1":
                    p = 0
                elif str(v)==str(L.base_ring().gen(0)):
                    p = 1
                elif "/sqrt" in str(v):
                    p=-1/2
                elif "sqrt" in str(v):
                    p=1/2
                else:
                    p = QQ(str(v)[2:].replace(")","").replace("(",""))
                if pmin==None or p<pmin:
                    pmin = p
            correction = 1/L.base_ring().gen(0)**ceil(pmin)
            self._correction = correction
        return self._correction
    
    @property
    def S1(self):
        return self._S1
    @property
    def S2(self):
        return self._S2
    
    @property
    def basepoint(self):
        return self._basepoint

    @property
    def critical_values(self):
        if not hasattr(self, "_critical_values"):
            self._critical_values =[c for c in Util.remove_duplicates(self.S1.critical_values + self.S2.critical_values) if c != "infinity"]
        return self._critical_values

    @property
    def fundamental_group(self):
        if not hasattr(self,'_fundamental_group'):
            fundamental_group = FundamentalGroupVoronoi([e for e in self.critical_values if e!= 'infinity'], self.basepoint)
            fundamental_group.sort_loops()
            self._critical_values = fundamental_group.points[1:]
            self._fundamental_group = fundamental_group
        return self._fundamental_group
    
    @property
    def paths(self):
        if not hasattr(self, '_paths'):
            self._paths = self.fundamental_group.pointed_loops_vertices
        return self._paths

    @property
    def thimbles(self):
        return self.monodromy_representation.thimbles

    @property
    def permuting_cycles(self):
        return self.monodromy_representation.permuting_cycles
    
    @property
    def extensions(self):
        """Representants of the extensions of the elliptic surface."""
        return self.monodromy_representation.extensions

    @property
    def borders_of_thimbles(self):
        return self.monodromy_representation.borders_of_thimbles

    @property
    def infinity_loops(self):
        """The linear combinations of thimbles that correspond to extensions along the (trivial) loop around infinity."""
        return self.monodromy_representation.infinity_loops
        
    @property
    def vanishing_cycles(self):
        if not hasattr(self, "_vanishing_cycles"):
            self._vanishing_cycles = [[(M-1)*v for v in permuting_cycles_of_M] for permuting_cycles_of_M, M in zip(self.permuting_cycles, self.monodromy_matrices)]
        return self._vanishing_cycles

    @property
    def infinity_loops(self):
        """The linear combinations of thimbles that correspond to extensions along the (trivial) loop around infinity."""
        return self.monodromy_representation.infinity_loops
    
    def lift(self, v):
        return self.monodromy_representation.resolve(self.monodromy_representation.lift(self.monodromy_representation.desingularise(v)))

    @property
    def homology(self):
        return self.monodromy_representation.homology
    
    def integrate(self, L):
        logger.info("Computing numerical transition matrices of operator of order %d and degree %d (%d edges total)."% (L.order(), L.degree(), len(self.fundamental_group.edges)))
        begin = time.time()

        integrator = Integrator(self.fundamental_group, L, self.ctx.nbits)
        transition_matrices = integrator.transition_matrices
        
        end = time.time()
        duration_str = time.strftime("%H:%M:%S",time.gmtime(end-begin))
        logger.info("Integration finished -- total time: %s."% (duration_str))

        return transition_matrices
    
    def cyclic_form_and_vectors(self):
        L1s = [self.S1.family.picard_fuchs_equation(v) for v in identity_matrix(2)]
        L2s = [self.S2.family.picard_fuchs_equation(v) for v in identity_matrix(2)]
        found = False
        for i1, L1 in enumerate(L1s):
            for i2, L2 in enumerate(L2s):
                Ltot = L1.symmetric_product(L2)
                if Ltot.order() ==4:
                    return self.S1.family.basis[i1], self.S1.family.basis[i2], Ltot

    @property
    def monodromy_matrices(self):
        if not hasattr(self,'_monodromy_matrices'):
            logger.info("Computing monodromy matrices")
            w1, w2, Ltot = self.cyclic_form_and_vectors()
            w1, w2 = self.S1.P.parent()(w1), self.S2.P.parent()(w2)

            Dt, = Ltot.parent().gens()
            
            transition_matrices = self.integrate(Ltot*Dt)
            
            d1 = derivatives_coords(self.S1, w1, 3)
            d2 = derivatives_coords(self.S2, w2, 3)
            derivatives_tensor = matrix([vector([0]*4)] + [derivative_of_product(d1, d2, i) for i in range(4)])
            
            pM = tens1(self.S1.fibre.period_matrix)*tens2(self.S2.fibre.period_matrix)
            
            integration_correction = diagonal_matrix([1/ZZ(factorial(k)) for k in range(5)])
            init_cond = (integration_correction * derivatives_tensor).submatrix(1,0)
            
            monodromy_matrices = [pM.inverse() * init_cond.inverse() * M.submatrix(1,1) * init_cond * pM for M in transition_matrices]
            monodromy_matrices = [M.change_ring(ZZ) for M in monodromy_matrices]


            Mtot = identity_matrix(4)
            for M in monodromy_matrices:
                Mtot = M * Mtot
            if Mtot != identity_matrix(4):
                self._critical_values = self.critical_values + ["infinity"]

                transition_matrix_infinity = 1
                for M in transition_matrices:
                    transition_matrix_infinity = M * transition_matrix_infinity
                transition_matrices += [transition_matrix_infinity.inverse()]
                
                monodromy_matrices += [(Mtot.inverse()).change_ring(ZZ)]
                
                path_infinity = -sum(self.paths)
                self._paths += [path_infinity]

            self._monodromy_matrices = monodromy_matrices
        return self._monodromy_matrices

    @property
    def monodromy_matrices_morsification(self):
        return [[Ms[2*i]*Ms[2*i+1] for i in range(len(Ms)/2)] for Ms in self.monodromy_representation.monodromy_matrices_desingularisation]
    
    @property
    def monodromy_representation(self):
        if not hasattr(self,'_monodromy_representation'):
            monodromy_matrices = self.monodromy_matrices
            expected_types = []
            for c in self.critical_values:
                if c in self.S1.critical_values:
                    expected_types += [self.S1.types[self.S1.critical_values.index(c)]]
                else:
                    expected_types += ["I0"]
            fibre_intersection_product = tens1(self.S1.fibre.intersection_product)*tens2(self.S2.fibre.intersection_product)
            self._monodromy_representation = MonodromyRepresentationFibreProduct(monodromy_matrices, fibre_intersection_product, expected_types)
        return self._monodromy_representation

    @property
    def cohomology(self):
        """a cohomology class is represented by `(w1, w2, denom)` where `w1` is a form of `self.S1`, `w2` is a form of `self.S2` and `denom` is a rational function of the parameter."""
        if not hasattr(self,"_cohomology"):
            S = self.S1.P.parent()
            R = S.base_ring()
            
            Swithu = PolynomialRing(R, ['t', 'u'])
            t,u = Swithu.gens() 
            phi = S.hom([t*u])
            phii = Swithu.hom([S.gens()[0], 1])
            f1 = Swithu(1)/phi(self.S1.P) # f1 is the rational function of which the residue is the holomorphic form of S1
            
            derivatives = [f1]
            for i in range(3):
                derivatives += [derivatives[-1].derivative(u)]
            nums = [phii(phi(self.S1.P**(i+1))*w)/factorial(i) for i, w in enumerate(derivatives)]

            w2 = vector([1,0])
            w1s = [S(Swithu(n)) for n in nums]
            w1s, denom = self.S1.family.coordinates(w1s)

            self._cohomology = [(w1, w2, denom) for w1 in w1s]
        return self._cohomology

    @property
    def picard_fuchs_equations(self):
        if not hasattr(self,"_picard_fuchs_operators"):
            picard_fuchs_equations = []
            for w1, w2, denom in self.cohomology:
                L1 = self.S1.family.picard_fuchs_equation(w1)
                L2 = self.S2.family.picard_fuchs_equation(w2)
                L = L1.symmetric_product(L2)*denom
                L = L*(1/self.correction)
                mul = lcm([c.denominator() for c in L.coefficients()])
                L = mul*L
                L = L.change_ring(L.base_ring().ring_of_integers())
                picard_fuchs_equations += [L]
            self._picard_fuchs_equations = picard_fuchs_equations
        return self._picard_fuchs_equations
    
    @property
    def transition_matrices(self):
        if not hasattr(self, '_transition_matrices'):
            transition_matrices = []
            for L in self.picard_fuchs_equations:
                L = L* L.parent().gens()[0]
                transition_matrices += [self.integrate(L)]
                if "infinity" in self.critical_values:
                    transition_matrix_infinity = 1
                    for M in transition_matrices[-1]:
                        transition_matrix_infinity = M*transition_matrix_infinity
                    transition_matrices[-1] += [transition_matrix_infinity**(-1)]
            self._transition_matrices = transition_matrices
        return self._transition_matrices
    
    @property
    def integrated_thimbles(self):
        if not hasattr(self, '_integrated_thimbles'):
            integration_correction = diagonal_matrix([1/ZZ(factorial(k)) for k in range(5)])
            pM = tens1(self.S1.fibre.period_matrix)*tens2(self.S2.fibre.period_matrix)

            _integrated_thimbles_all = []

            for transition_matrices, w in zip(self.transition_matrices, self.cohomology):
                w1, w2, denom = w

                numerator1 = (w1[0]*self.S1.P + w1[1]*self.S1.family.pol.parent()(self.S1.family.coho1.basis()[1]))*self.correction/denom
                numerator2 = (w2[0]*self.S2.P + w2[1]*self.S2.family.pol.parent()(self.S2.family.coho1.basis()[1]))
                d1 = derivatives_coords(self.S1, numerator1, 3)
                d2 = derivatives_coords(self.S2, numerator2, 3)
                derivatives_tensor = matrix([vector([0]*4)] + [derivative_of_product(d1, d2, i) for i in range(4)])
                initial_conditions = integration_correction * derivatives_tensor * pM

                _integrated_thimbles = []
                for i, ps in enumerate(self.permuting_cycles):
                    for p in ps:
                        _integrated_thimbles += [(transition_matrices[i] * initial_conditions.submatrix(0,0,transition_matrices[i].nrows()) * p)[0]]
                _integrated_thimbles_all += [_integrated_thimbles]
            self._integrated_thimbles = _integrated_thimbles_all
        return self._integrated_thimbles
    
    @property
    def primary_lattice(self):
        return self.monodromy_representation.primary_lattice
    
    @property
    def primary_periods(self):
        if not hasattr(self, "_primary_periods"):
            homology_mat = matrix(self.extensions).transpose()
            integrated_thimbles =  matrix(self.integrated_thimbles)
            self._primary_periods = integrated_thimbles * homology_mat
        return self._primary_periods
    

    @property
    def components_of_singular_fibres(self):
        return self.monodromy_representation.components_of_singular_fibres

    @property
    def period_matrix(self):
        if not hasattr(self, '_period_matrix'):
            periods_tot = block_matrix([[self.primary_periods, zero_matrix(len(self.cohomology), len(flatten(self.components_of_singular_fibres)))]])
            
            res = self.primary_lattice.solve_right(matrix([self.monodromy_representation.proj(v) for v in self.homology]).transpose())
            self._period_matrix = periods_tot * res
        return self._period_matrix
    
    @property
    def intersection_product(self):
        return self.monodromy_representation.intersection_product_resolution
    

    @property
    def types(self):
        return self.monodromy_representation.types


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from sage.plot.plot import list_plot


from sage.modules.free_module_element import vector
from sage.matrix.constructor import matrix

from .util import Util

class PointedLoop(object):
    def __init__(self, path):
        assert path[0] == path[-1], "given path is not a loop"
        self.path = Util.simplify_path(path)
    
    def __str__(self):
        s = str(self.path)
        if len(s)>=100:
            s = s[:50] + "..." + s[-50:]
        return "loop pointed at " + str(self.path[0]) +" along " + s
    def __repr__(self):
        s = str(self.path)
        if len(s)>=100:
            s = s[:50] + "..." + s[-50:]
        return "loop pointed at " + str(self.path[0]) +" along " + s
    
    def __add__(self, other):
        return PointedLoop(self.path + other.path)
    def __radd__(self, other):
        if other==0:
            return self
        return PointedLoop(self.path + other.path)
        
    def __sub__(self, other):
        return PointedLoop(self.path + list(reversed(other.path)))

    def __neg__(self):
        return PointedLoop(list(reversed(self.path)))
    
    def __rmul__(self, other):
        if other ==0:
            return PointedLoop([self.path[0]])
        if other >0:
            return PointedLoop(self.path * other)
        if other <0:
            return PointedLoop(list(reversed(self.path)) * -other)
        
    def __mul__(self, other):
        if other ==0:
            return PointedLoop([self.path[0]])
        if other >0:
            return PointedLoop(self.path * other)
        if other <0:
            return PointedLoop(list(reversed(self.path)) * -other)
    
    def __iter__(self):
        return iter(self.path)

    def draw(self, basepoint_free=False, **kwds):
        if not basepoint_free:
            return list_plot([[c.real(), c.imag()] for c in self.path], True, **kwds)
            
        path = self.path[len(self.path)//2:] + self.path[:len(self.path)//2]
        path = Util.simplify_path(path) + [path[0]]
        return list_plot([[c.real(), c.imag()] for c in path], True, **kwds)
    
    def __getitem__(self, i):
        return self.path[i]
    
    def __len__(self):
        return len(self.path)

    @property
    def edges(self):
        edges = []
        for i in range(len(self.path)-1):
            edges += [self.path[i:i+2]]
        return edges
    
def simplify_conjugation(path_indices):
    res = path_indices.copy()
    i=0
    while i<len(res)-1:
        if res[i][0] == res[i+1][0]:
            newp = res[i][1]+res[i+1][1]
            if newp==0:
                res = res[:i]+res[i+2:]
                i=max(i-1,0)
            else:
                res = res[:i]+[(res[i][0], newp)]+res[i+2:]
        else:
            i+=1
    return res

class LoadedPointedLoop(object):
    def __init__(self, fibration, letters):
        self.letters = simplify_conjugation(letters)
        self.fibration = fibration

    def __add__(self, other):
        if other == 0:
            return self
        return LoadedPointedLoop(self.fibration, self.letters + other.letters)
    def __radd__(self, other):
        return self.__add__(other)
        
    def __sub__(self, other):
        return self.__add__(-other)
        
    def __neg__(self):
        newletters = []
        for (i,p) in self.letters:
            newletters = [(i,-p)] + newletters
        return LoadedPointedLoop(self.fibration, newletters)
        
       
    def conjugate(self, other):
        return other + self - other

    def __repr__(self):
        s = str(self.letters)
        if len(s)>200:
            s = s[:50] + "<...>" + s[-50:]
        return "loaded loop with representation " + s
        
    def __str__(self):
        s = str(self.letters)
        if len(s)>200:
            s = s[:50] + "<...>" + s[-50:]
        return "loaded loop with representation " + s
    
    def __len__(self):
        return len(self.letters)
        
    def __getitem__(self, index):
        if isinstance(index, slice):
            # Return a new Foo with the sliced list
            return LoadedPointedLoop(self.fibration, self.letters[index])
        else:
            # Return a single item
            return self.letters[index]

    def __iter__(self):
        return iter(self.letters)

    def __contains__(self, item):
        return item in self.letters

    def index(self, item):
        return self.letters.index(item)
        
    @property
    def monodromy_matrix(self):
        M = 1
        for (i, p) in self.letters:
            M = self.fibration.monodromy_matrices[i]**p * M
        return M

    def extend(self, gamma):
        num_perm = [len(permlist) for permlist in self.fibration.permuting_cycles]
        res = vector([0 for i in range(sum(num_perm))])
        for (i,p) in self.letters:
            e = 1 if p>0 else -1
            for k in range(abs(p)):
                v = (self.fibration.monodromy_matrices[i]**e-1)*gamma
                decomp = matrix([(self.fibration.monodromy_matrices[i]-1)*per for per in self.fibration.permuting_cycles[i]]).solve_left(v)
                start = sum(num_perm[:i])
                for j, c in enumerate(decomp):
                    res[start+j] += c
                gamma = self.fibration.monodromy_matrices[i]**e*gamma
        return res

    @property
    def path(self):
        return sum([p*self.fibration.paths[i] for i,p in self.letters])
        
    @property
    def indices_inside(self):
        totals = [0 for i in range(len(self.fibration.critical_values))]
        for i, p in self.letters:
            totals[i] += p
        assert all([t in [0,1] for t in totals]), "loop is not enclosing a region"
        return [i for i,t in enumerate(totals) if t==1]
    
    def draw(self, **kwds):
        return self.path.draw(**kwds)
    


class HomotopyRepresentation(object):
    def __init__(self, fibration, loaded_paths=None):
        if loaded_paths == None:
            indices = [[(i,1)] for i in range(len(fibration.critical_values))]
            loaded_paths = [LoadedPointedLoop(fibration, i) for i in indices]
        self.loaded_paths = loaded_paths
        self.fibration = fibration

    def __repr__(self):
        s = "homotopy representation with " + str(len(self.loaded_paths)) + " loops"
        for l in self.loaded_paths:
            s += "\n  -  " + l.__str__()
        return s
        
    @property
    def monodromy_representation(self):
        return [l.monodromy_matrix for l in self.loaded_paths]

    def act_by_braid(self, i, p):
        assert i<len(self.loaded_paths)-1, "braid index is too high"
        indices = self.loaded_paths[:i]
        if p==1:
            indices += [self.loaded_paths[i+1].conjugate(self.loaded_paths[i]), self.loaded_paths[i]]
        elif p==-1:
            indices += [self.loaded_paths[i+1],self.loaded_paths[i].conjugate(-self.loaded_paths[i+1])]
        else:
            raise Exception("braid power i unknown")
        indices += self.loaded_paths[i+2:]
        return HomotopyRepresentation(self.fibration, loaded_paths=indices)

    def act_by_braids(self, list_of_braids):
        res = self
        for i, n in list_of_braids:
            p = -1 if n<0 else 1
            for j in range(abs(n)):
                res = res.act_by_braid(i, p)
        return res

    @property
    def length(self):
        return sum([len(l) for l in self.loaded_paths])
    
def homotopy_representation_from_allowable_loop(allowable_loop):
    fibration = allowable_loop.fibration
    inside_points = allowable_loop.indices_inside
    n = allowable_loop.index((inside_points[0],1))
    AL = allowable_loop
    done = []
    for k in inside_points:
        n = AL.index((k,1))
        c = AL[:n]
        done += [AL[n:n+1].conjugate(c)]
        AL = c + AL[n+1:]
    AL = - allowable_loop + sum(HomotopyRepresentation(allowable_loop.fibration).loaded_paths)
    for k in range(len(fibration.critical_values)):
        if k in inside_points:
            continue
        n = AL.index((k,1))
        c = AL[:n]
        done += [AL[n:n+1].conjugate(c)]
        AL = c + AL[n+1:]
    return HomotopyRepresentation(fibration, done)


# -*- coding: utf-8 -*-

# lefschetz-family
# Copyright (C) 2021  Eric Pichon-Pharabod

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sage.all

from ore_algebra import *

from sage.modules.free_module_element import vector
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.matrix.constructor import matrix
from sage.rings.integer_ring import ZZ
from sage.matrix.special import identity_matrix
from sage.matrix.special import diagonal_matrix
from sage.matrix.special import block_matrix
from sage.matrix.special import block_diagonal_matrix
from sage.matrix.special import zero_matrix

from sage.functions.other import floor
from sage.functions.other import ceil
from sage.arith.misc import gcd
from sage.misc.misc_c import prod
from sage.misc.flatten import flatten


from sage.symbolic.relation import solve
from sage.symbolic.ring import SR

from .monodromyRepresentation import MonodromyRepresentation
from .util import Util


import logging
import time

logger = logging.getLogger(__name__)

def tens1(M):
    return block_diagonal_matrix([M, M], subdivide=False)
def tens2(M):
    return block_matrix(list(map(lambda r:list(map(lambda c: c*identity_matrix(M.nrows()), r)), M.rows())), subdivide=False)

def tens11(M):
    return block_diagonal_matrix([M, identity_matrix(2)], subdivide=False)
def tens12(M):
    return block_diagonal_matrix([identity_matrix(2), M], subdivide=False)
def tens21(M):
    res = identity_matrix(4)
    for i in range(2):
        for j in range(2):
            res[2*i,2*j] = M[i,j]
    return res
def tens22(M):
    res = identity_matrix(4)
    for i in range(2):
        for j in range(2):
            res[2*i+1,2*j+1] = M[i,j]
    return res

class MonodromyRepresentationFibreProduct(MonodromyRepresentation):

    def __init__(self, monodromy_matrices, intersection_product, expected_types):
        super().__init__(monodromy_matrices, intersection_product)
        self.expected_types = expected_types

    @property
    def types(self):
        if not hasattr(self, "_types"):
            types = []
            for M, expected_type in zip(self.monodromy_matrices, self.expected_types):
                M1, M2 = self.disentangle(M)
                t, _, n = self.monodromy_class(M1)
                if t + str(n) != expected_type:
                    M1, M2 = -M1, -M2
                    t, _, n = self.monodromy_class(M1)
                assert t + str(n) == expected_type, "Unexpected fibre type"
                type1 = t + str(n) if t in ["I", "I*"] else t
                t, _, n = self.monodromy_class(M2)
                type2 = t + str(n) if t in ["I", "I*"] else t
                types += [type1 + " x " + type2]
            self._types = types
        return self._types
    
    def disentangle(self, M):
        M1 = M.submatrix(0,0,2,2)
        if M1 ==0:
            M1 = M.submatrix(0,2,2,2)
        M1 = (M1 / M1.det()**(1/2)).change_ring(ZZ)
        M2 = matrix(2, [(M.submatrix(i,j,2,2)*M1.inverse())[0,0] for i,j in [[0,0],[0,2],[2,0],[2,2]]])
        return M1, M2
    
    def desingularise_matrix(self, M, expected_type):
        M1, M2 = self.disentangle(M)
        ty, base_change, nu = self.monodromy_class(M1)
        
        if ty + str(nu) != expected_type:
            M1, M2 = -M1, -M2
            ty, base_change, nu = self.monodromy_class(M1)

        decomposition = self.fibre_confluence[ty]

        mats = []
        for M in decomposition[:-1]:
            mats += [tens11(base_change * M * base_change.inverse())]
            mats += [tens12(base_change * M * base_change.inverse())]
        mats += [tens11(base_change * decomposition[-1] * base_change.inverse()), tens12(base_change * decomposition[-1] * base_change.inverse())] * nu
        ty, base_change, nu = self.monodromy_class(M2)
        decomposition = self.fibre_confluence[ty]
        for M in decomposition[:-1]:
            mats += [tens21(base_change * M * base_change.inverse())]
            mats += [tens22(base_change * M * base_change.inverse())]
        mats += [tens21(base_change * decomposition[-1] * base_change.inverse()), tens22(base_change * decomposition[-1] * base_change.inverse())] * nu

        mats = [M.change_ring(ZZ) for M in mats]
        return mats
    
    @property
    def monodromy_matrices_desingularisation(self):
        if not hasattr(self, '_monodromy_matrices_desingularisation'):
            monodromy_matrices_desingularisation = []
            for M, expected_type in zip(self.monodromy_matrices, self.expected_types):
                decomposition = self.desingularise_matrix(M, expected_type)
                assert prod(list(reversed(decomposition))) == M
                monodromy_matrices_desingularisation += [decomposition]
            self._monodromy_matrices_desingularisation = monodromy_matrices_desingularisation
        return self._monodromy_matrices_desingularisation


    def _compute_intersection_product_thimbles(self,i,j):
        vi = self.permuting_cycles_desingularisation[i]
        vj = self.permuting_cycles_desingularisation[j]

        Mi = flatten(self.monodromy_matrices_desingularisation)[i]
        Mj = flatten(self.monodromy_matrices_desingularisation)[j]
        
        di, dj = (Mi-1) * vi, (Mj-1) * vj

        res = di * self.fibre_intersection_product * dj
        resid = -vi * self.fibre_intersection_product * dj

        loopi = i//2
        loopj = j//2

        if loopi==loopj:
            return resid
        if loopi<loopj:
            return res
        else:
            return 0

    @property
    def intersection_product_resolution(self):
        if not hasattr(self, "_intersection_product_resolution"):
            cycles = matrix(self.extensions_resolution)
            self._intersection_product_resolution = cycles * self.intersection_product * cycles.transpose()
        return self._intersection_product_resolution

    @property
    def vanishing_cycles_conifold_transition(self):
        if not hasattr(self, "_vanishing_cycles_conifold_transition"):
            # first we figure out which pairs of singular components are there
            ords = []
            for Ms in self.monodromy_matrices_desingularisation:
                managed=False
                for i, M in enumerate(Ms):
                    if M.submatrix(2,0,2,2)!=0 or M.submatrix(0,2,2,2)!=0:
                        ords +=[i]
                        managed=True
                        break
                if not managed:
                    ords += [len(Ms)]
            # then we pair the vanishing cycles accordingly
            ranktot = 0
            rankmax = sum([len(Ms) for Ms in self.monodromy_matrices_desingularisation])
            sing_comps = []
            for n, M in zip(ords, self.vanishing_cycles_desingularisation):
                M = matrix(M)
                rank = M.dimensions()[0]
                vanishing = []
                Ni = M[:n].nrows()//2
                Nj = M[n:].nrows()//2
                for i in range(Ni):
                    for j in range(Nj):
                        vc = M[:n][2*i:2*i+2].stack(M[n:][2*j:2*j+2]).left_kernel_matrix().row(0)
                        vc = vector([0]*2*i + list(vc[:2]) + [0]*2*(Ni-i-1) + [0]*2*j +  list(vc[2:]) + [0]*2*(Nj-j-1))
                        vanishing += [vc]
                sing_comps += [[vector([0]*ranktot+list(v) + [0]*(rankmax-ranktot-rank)) for v in vanishing]]
                ranktot+=rank
            self._vanishing_cycles_conifold_transition = sing_comps
        return self._vanishing_cycles_conifold_transition
    
    def resolve(self, v):
            """Given a combination of thimbles of desingularisation, gives the corresponding homology class"""
            vccts = [self.lift(v2) for v2 in flatten(self.vanishing_cycles_conifold_transition)]
            assert matrix(vccts) * self.intersection_product * v ==0, "cycle does not survive resolution"
            v = matrix(self.extensions_resolution + vccts).solve_left(v)
            return vector(list(v)[:-len(vccts)])
    
    def proj(self, v):
            """Given a resolved homology class, gives a corresponding homology class (up to a conifold transition vanishing cycle)"""
            return v * matrix(self.extensions_resolution)
    
    @property
    def extensions_resolution(self):
        if not hasattr(self, "_extensions_resolution"):
            vanishing_cycles = matrix([self.lift(v) for v in flatten(self.vanishing_cycles_conifold_transition)])
            vanishing_cycles_orth = (vanishing_cycles * self.intersection_product).change_ring(ZZ).right_kernel_matrix()
            toremove = vanishing_cycles.image().intersection(vanishing_cycles_orth.image()).basis_matrix()
            compl = Util.find_complement(vanishing_cycles_orth.solve_left(toremove).change_ring(ZZ))
            
            self._extensions_resolution = (compl * vanishing_cycles_orth).rows()
        return self._extensions_resolution

    @property
    def homology(self): 
        if not hasattr(self, '_homology'):
            self._homology = identity_matrix(len(self.extensions_resolution)).rows()
        return self._homology

    @property
    def primary_lattice(self):
        if not hasattr(self, '_primary_lattice'):
            extensions = [super(MonodromyRepresentationFibreProduct, self).lift(self.desingularise(v)) for v in self.extensions]
            # extensions = matrix(extensions).change_ring(ZZ).image().basis()
            components_of_singular_fibres = [super(MonodromyRepresentationFibreProduct, self).lift(v) for v in flatten(self.components_of_singular_fibres, max_level=2)]
            primary_lattice = extensions + components_of_singular_fibres
            self._primary_lattice = matrix(primary_lattice).transpose()
        return self._primary_lattice
    
    @property
    def primary_periods(self):
        if not hasattr(self, '_primary_periods'):
            homology_mat = matrix(self.extensions).transpose()
            integrated_thimbles =  matrix(self.integrated_thimbles)
            self._primary_periods = integrated_thimbles * homology_mat
        return self._primary_periods


    
    @property
    def permuting_cycles(self):
        if not hasattr(self, '_permuting_cycles'):
            permuting_cycles = [identity_matrix(self.dim).rows() for i in range(len(self.monodromy_matrices))]
            self._permuting_cycles = permuting_cycles
        return self._permuting_cycles
    
    @property
    def infinity_loops(self):
        """The linear combinations of thimbles that correspond to extensions along the (trivial) loop around infinity."""
        if not hasattr(self, '_infinity_loops'):
            infinity_cycles = []
            for i in range(self.dim):
                v = vector([1 if k==i else 0 for k in range(self.dim)])
                coefs = []
                for j in range(len(self.monodromy_matrices)):
                    M = self.monodromy_matrices[j]
                    if len(self.permuting_cycles[j])==0:
                        continue
                    coefs += list(v)
                    v = self.monodromy_matrices[j]*v
                infinity_cycles+=[vector(coefs)]
            self._infinity_loops = matrix(infinity_cycles).change_ring(ZZ).rows()
        return self._infinity_loops
    
    @property
    def add(self):
        if not hasattr(self, '_add'):
            self._add = 0
        return self._add

    classes = {
        "I":matrix([[1,1],[0,1]]),
        "II":matrix([[1,1],[-1,0]]),
        "III":matrix([[0,1],[-1,0]]),
        "IV":matrix([[0,1],[-1,-1]]),
        "I*":matrix([[-1,-1],[0,-1]]),
        "II*":matrix([[0,-1],[1,1]]),
        "III*":matrix([[0,-1],[1,0]]),
        "IV*":matrix([[-1,-1],[1,0]]),
    }

    U = matrix([[1,1],[0,1]])
    V = matrix([[1,0],[-1,1]])

    fibre_confluence = {
        "I": [U],
        "II": [U, V],
        "III": [V,U,V],
        "IV": [U,V]*2,
        "I*": [U,V]*3+[U],
        "II*": [U,V]*5,
        "III*": [U,V]*3+[V,U,V],
        "IV*": [U,V]*4
    }

    @classmethod
    def monodromy_class(cls, M):
        M=M.change_ring(ZZ)
        if M.trace()==2:
            if M==identity_matrix(2):
                return "I", identity_matrix(2), 0
            else:
                return "I", cls.normalize_Iv(M)[0], cls.normalize_Iv(M)[1]
        if M.trace()==1:
            if M[0,1]>0:
                return "II", cls.find_conjugation(M, cls.classes["II"])[0], 1
            else:
                return "II*", cls.find_conjugation(M, cls.classes["II*"])[0], 1
        if M.trace()==0:
            try:
                bc = cls.find_conjugation(M, cls.classes["III"])[0]
                return "III", bc, 1
            except:
                bc = cls.find_conjugation(M, cls.classes["III*"])[0]
                return "III*", bc, 1
        if M.trace()==-1:
            try:
                bc = cls.find_conjugation(M, cls.classes["IV"])[0]
                return "IV", bc, 1
            except:
                bc = cls.find_conjugation(M, cls.classes["IV*"])[0]
                return "IV*", bc, 1
        if M.trace()==-2:
            if M==-identity_matrix(2):
                return "I*", identity_matrix(2), 0
            else:
                return "I*", cls.normalize_Ivstar(M)[0], cls.normalize_Ivstar(M)[1]
    

    @classmethod
    def normalize_Iv(cls, M):
        vanishing = (M-1).transpose().image().basis()[0]
        v = gcd(vanishing)
        v = ZZ(gcd(vanishing))
        vanishing = vanishing/v
        vanishing = vanishing.change_ring(ZZ)
        D, U, V = (M-1).smith_form()
        compl = V*vector([1,0])
        if (M-1)*compl != vanishing*v:
            compl = -compl
        assert (M-1)*compl == vanishing*v, "could not find permuting vector"
        bc = matrix([vanishing, compl]).transpose()
        if bc.det()!=1:
            bc = matrix([-vanishing, compl]).transpose()
        assert bc.det()==1, "could not find permuting vector"
        return bc, v
    
    @classmethod
    def normalize_Ivstar(cls, M):
        vanishing = (M+1).transpose().image().basis()[0]
        v = ZZ(gcd(vanishing))
        vanishing = vanishing/v
        vanishing = vanishing.change_ring(ZZ)
        D, U, V = (M+1).smith_form()
        compl = V*vector([1,0])
        if (M+1)*compl != vanishing*v:
            compl = -compl
        assert (M+1)*compl == vanishing*v, "could not find permuting vector"
        bc = matrix([vanishing, compl]).transpose()
        if bc.det()!=1:
            bc = matrix([-vanishing, compl]).transpose()
        assert bc.det()==1, "could not find permuting vector"
        return bc, v
    
    @classmethod
    def find_conjugation(cls, M1, M2):
        R = PolynomialRing(ZZ, ['a','b','c','d'])
        [a,b,c,d] = R.gens()
        bci = matrix([[a,b],[c,d]])
        eqs = (bci*M1 - M2*bci).coefficients() + [a*d-b*c-1]
        I = R.ideal(eqs)
        eq = I.elimination_ideal([a,b]).gens()[0]
        
        x = SR('x')
        discr = eq.coefficient(c)**2 - 4*eq.coefficient(c**2)*eq(c=0)
        possible_ds = range(ceil(solve(discr(d=x)>0,x)[0][0].right_hand_side()), floor(solve(discr(d=x)>0,x)[0][1].right_hand_side())+1)
        Qt = PolynomialRing(ZZ, 't')
        t=Qt.gens()[0]
        
        sols=[]
        for dv in possible_ds:
            possible_cs = eq(d=dv, c=t).roots(multiplicities=False)
            if len(possible_cs)>0:
                for cv in possible_cs:
                    I2 = R.ideal(eqs+[c-cv, d-dv])
                    av = -I2.elimination_ideal([b,c,d]).gens()[0](a=0)
                    bv = -I2.elimination_ideal([a,c,d]).gens()[0](b=0)
                    sols += [matrix([[av,bv],[cv,dv]])]    
        return [(s**-1).change_ring(ZZ) for s in sols]